@@ -79,6 +79,7 @@ struct resolv_lookup {
 struct resolv_query {
 	GResolv *resolv;
 
+	int nr_ns;
 	guint timeout;
 
 	uint16_t msgid;
@@ -471,6 +472,8 @@ static void sort_and_return_results(struct resolv_lookup *lookup)
 	char buf[INET6_ADDRSTRLEN + 1];
 	GResolvResultStatus status;
 	char **results = g_try_new0(char *, lookup->nr_results + 1);
+	GResolvResultFunc result_func = lookup->result_func;
+	void *result_data = lookup->result_data;
 	int i, n = 0;
 
 	if (!results)
@@ -494,7 +497,7 @@ static void sort_and_return_results(struct resolv_lookup *lookup)
 		} else
 			continue;
 
-		results[n++] = strdup(buf);
+		results[n++] = g_strdup(buf);
 	}
 
 	results[n++] = NULL;
@@ -513,11 +516,11 @@ static void sort_and_return_results(struct resolv_lookup *lookup)
 	debug(lookup->resolv, "lookup %p received %d results", lookup, n);
 
 	g_queue_remove(lookup->resolv->lookup_queue, lookup);
+	destroy_lookup(lookup);
 
-	lookup->result_func(status, results, lookup->result_data);
+	result_func(status, results, result_data);
 
 	g_strfreev(results);
-	destroy_lookup(lookup);
 }
 
 static gboolean query_timeout(gpointer user_data)
@@ -537,12 +540,11 @@ static gboolean query_timeout(gpointer user_data)
 	}
 
 	g_queue_remove(resolv->query_queue, query);
+	destroy_query(query);
 
 	if (lookup->ipv4_query == NULL && lookup->ipv6_query == NULL)
 		sort_and_return_results(lookup);
 
-	destroy_query(query);
-
 	return FALSE;
 }
 
@@ -578,12 +580,13 @@ static void flush_nameservers(GResolv *resolv)
 static int send_query(GResolv *resolv, const unsigned char *buf, int len)
 {
 	GList *list;
+	int nr_ns;
 
 	if (resolv->nameserver_list == NULL)
 		return -ENOENT;
 
-	for (list = g_list_first(resolv->nameserver_list);
-					list; list = g_list_next(list)) {
+	for (list = g_list_first(resolv->nameserver_list), nr_ns = 0;
+				list; list = g_list_next(list), nr_ns++) {
 		struct resolv_nameserver *nameserver = list->data;
 		int sk, sent;
 
@@ -597,7 +600,7 @@ static int send_query(GResolv *resolv, const unsigned char *buf, int len)
 			continue;
 	}
 
-	return 0;
+	return nr_ns;
 }
 
 static gint compare_lookup_id(gconstpointer a, gconstpointer b)
@@ -664,6 +667,11 @@ static void parse_response(struct resolv_nameserver *nameserver,
 
 	ns_initparse(buf, len, &msg);
 
+	list = g_queue_find_custom(resolv->query_queue,
+			GUINT_TO_POINTER(ns_msg_id(msg)), compare_query_msgid);
+	if (!list)
+		return;
+
 	rcode = ns_msg_getflag(msg, ns_f_rcode);
 	count = ns_msg_count(msg, ns_s_an);
 
@@ -694,21 +702,15 @@ static void parse_response(struct resolv_nameserver *nameserver,
 		break;
 	}
 
-	list = g_queue_find_custom(resolv->query_queue,
-			GUINT_TO_POINTER(ns_msg_id(msg)), compare_query_msgid);
-	if (!list)
-		return;
-
 	query = list->data;
+	query->nr_ns--;
+
 	lookup = query->lookup;
 
-	if (query == lookup->ipv6_query) {
+	if (query == lookup->ipv6_query)
 		lookup->ipv6_status = status;
-		lookup->ipv6_query = NULL;
-	} else if (query == lookup->ipv4_query) {
+	else if (query == lookup->ipv4_query)
 		lookup->ipv4_status = status;
-		lookup->ipv4_query = NULL;
-	}
 
 	for (i = 0; i < count; i++) {
 		ns_parserr(&msg, ns_s_an, i, &rr);
@@ -728,12 +730,19 @@ static void parse_response(struct resolv_nameserver *nameserver,
 		}
 	}
 
+	if (status != G_RESOLV_RESULT_STATUS_SUCCESS && query->nr_ns > 0)
+		return;
+
+	if (query == lookup->ipv6_query)
+		lookup->ipv6_query = NULL;
+	else
+		lookup->ipv4_query = NULL;
+
 	g_queue_remove(resolv->query_queue, query);
+	destroy_query(query);
 
 	if (lookup->ipv4_query == NULL && lookup->ipv6_query == NULL)
 		sort_and_return_results(lookup);
-
-	destroy_query(query);
 }
 
 static gboolean received_udp_data(GIOChannel *channel, GIOCondition cond,
@@ -975,7 +984,8 @@ static gint add_query(struct resolv_lookup *lookup, const char *hostname, int ty
 
 	debug(lookup->resolv, "sending %d bytes", len);
 
-	if (send_query(lookup->resolv, buf, len) < 0) {
+	query->nr_ns = send_query(lookup->resolv, buf, len);
+	if (query->nr_ns <= 0) {
 		g_free(query);
 		return -EIO;
 	}
@@ -1003,11 +1013,11 @@ guint g_resolv_lookup_hostname(GResolv *resolv, const char *hostname,
 {
 	struct resolv_lookup *lookup;
 
-	debug(resolv, "hostname %s", hostname);
-
 	if (resolv == NULL)
 		return 0;
 
+	debug(resolv, "hostname %s", hostname);
+
 	if (resolv->nameserver_list == NULL) {
 		int i;
 
